package org.processmining.indatabasepreprocessing.alignment;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * This class implements an idea derived from Andrea's paper in 
 * Online Conformance Checking using Behavioural Patterns.
 * Instead of using conformance metrics, I propose to compute alignment cost.
 * However I am not able yet to interpret the results.
 * Experiment shows a pretty big gap between normal alignment cost with this cost.
 * Note: pMin and pMax is coded manually. Once we benefit from the results, we can make it automatic.
 * 
 * @author asyamsiy
 *
 */
public class BehaviouralAlignment {
	
	Map<String, Set<String>> pMin;
	Map<String, Set<String>> pMax;
	Set<String> validPair;	
	Map<String, ArrayList<LogEvent>> log;
	Map<String, Integer> costMap;
	String last_elem;
	String last_sync;
	Set<String> obs;
	int cost;
	
	public static void main(String[] args) throws IOException, ParseException {
		BehaviouralAlignment ba = new BehaviouralAlignment();
		
		ba.transformLog(new File("C:\\Users\\asyamsiy\\Documents\\Experiment\\Alignment\\data\\log_behav_align_100.csv"));
		ba.createPMin();
		ba.createPMax();
		ba.getAlignmentCost();
		ba.printCost();
		ba.getFitnessValue(5);
	}
	
	private void getFitnessValue(int minimumLength) {
		int totalCost = 0;
		int totalEvents = 0;
		
		for(String caseid : costMap.keySet()) {
			totalCost += costMap.get(caseid);
		}
		for(String caseid : log.keySet()) {
			totalEvents += log.get(caseid).size();
		}
		
		double fitness = 1.0 - ((double) totalCost/ (double)(totalEvents + minimumLength));
		System.out.println("FITNESS: " + fitness);
		
	}
	
	private void getAlignmentCost() {
		costMap = new HashMap<String, Integer>();
		
		for(String caseid : log.keySet()) {
			ArrayList<LogEvent> events = log.get(caseid);
			last_elem = events.get(0).label;
			last_sync = events.get(0).label;
			obs = new HashSet<String>();
			cost = 0;
			
			for(int i = 0; i < events.size()-1; i++) {
				String a = events.get(i).label;
				String b = events.get(i+1).label;
				String pair = a + "-" + b;
				
				//System.out.println("a: " + a + " b: " + b + " last_sync: " + last_sync + " last_elem: " + last_elem);
				
				if(validPair.contains(pair) && last_sync.equals(last_elem)) {
					processValidPair(pair, b);				
				
				} else {
					
					if(last_elem.equals(last_sync)) {
						//System.out.println("test1: " + a + "-"  + b);
						cost++;
						last_elem = b;
					
					} else {
						if(validPair.contains(last_sync + "-"  + b)) {
							//System.out.println("test2: " + a + "-"  + b);
							last_elem = last_sync;
							processValidPair(last_sync + "-"  + b, b);
						} else {
							cost++;
							last_elem = b;
						}
					}					
				}
			}
			
			costMap.put(caseid, cost);
		}
	}
	
	private void processValidPair(String pair, String b) {
		if(obs.size() >= pMin.get(pair).size() && obs.size() <= pMax.get(pair).size()) {
			obs.add(pair);
			
		} else {
			Set<String> n = new HashSet<String>(pMin.get(pair));
			n.removeAll(obs);
			obs.addAll(n);
			
			if(last_elem.equals(last_sync)) {
				cost += n.size();
			} else {
				cost -= 1;
				cost += (n.size() - 1);
			}
		}
		
		last_elem = last_sync = b;					
	}
	
	private void printCost() {
		for(String caseid : costMap.keySet()) {
			System.out.println("case: " + caseid + "\t cost: " + costMap.get(caseid));
		}
	}
	
	private void transformLog(File path) throws IOException, ParseException {
		log = new HashMap<String, ArrayList<LogEvent>>();	
	
		BufferedReader in = new BufferedReader(new FileReader(path));
		String line = in.readLine(); 
		
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd hh:mm:ss.SSS");
		
		while((line = in.readLine()) != null) {
			String[] arr = line.split(",");
			String caseId = arr[0].substring(1, arr[0].length()-1);
			String activity = arr[1].substring(1, arr[1].length()-1);
			String timeStr = arr[2].substring(1,  arr[2].length()-1);
			Date parsedDate = dateFormat.parse(timeStr);
			Timestamp time = new java.sql.Timestamp(parsedDate.getTime());
			
			ArrayList<LogEvent> sequence = log.get(caseId);
			if (sequence == null){
				sequence = new ArrayList<LogEvent>();
				log.put(caseId, sequence);
			}
			//Add the event to the sequence that belongs to the case identifier of that event. 
			sequence.add(new LogEvent(time.getTime(), activity, ""));
			
		}
		
		// sort the log
		for (String caseId : log.keySet()) {
			Collections.sort(log.get(caseId));
		}

		in.close();
	}
	
	private void createPMin() {
		pMin = new HashMap<String, Set<String>>();
		
		Set<String> pMinSet;
		
		pMinSet = new HashSet<String>();
		pMin.put("A-A1", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1");
		pMin.put("A1-A2", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1"); pMinSet.add("A1-A2");
		pMin.put("A2-A1", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1");
		pMin.put("A1-B", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1"); pMinSet.add("A1-B");
		pMin.put("B-C", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1"); pMinSet.add("A1-B");
		pMin.put("B-E", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1"); pMinSet.add("A1-B"); pMinSet.add("B-C");
		pMin.put("C-D", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1"); pMinSet.add("A1-B"); pMinSet.add("B-C"); pMinSet.add("C-D");
		pMin.put("D-F", pMinSet);
		
		pMinSet = new HashSet<String>();
		pMinSet.add("A-A1"); pMinSet.add("A1-B"); pMinSet.add("B-E");
		pMin.put("E-F", pMinSet);
		
		validPair = pMin.keySet();
	}
	
	private void createPMax() {
		pMax = new HashMap<String, Set<String>>();
		
		Set<String> pMaxSet;
		
		pMaxSet = new HashSet<String>();
		pMax.put("A-A1", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("A1-A2", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("A2-A1", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("A1-B", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-B"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("B-C", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-B"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("B-E", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-B"); pMaxSet.add("B-C"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("C-D", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-B"); pMaxSet.add("B-C"); pMaxSet.add("C-D"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("D-F", pMaxSet);
		
		pMaxSet = new HashSet<String>();
		pMaxSet.add("A-A1"); pMaxSet.add("A1-B"); pMaxSet.add("B-E"); pMaxSet.add("A1-A2"); pMaxSet.add("A2-A1");
		pMax.put("E-F", pMaxSet);
		
	}
	
	class LogEvent implements Comparable<LogEvent> {
		long time;
		String label;
		String resource;
		
		LogEvent(long time, String label, String resource){
			this.time = time;
			this.label = label;
			this.resource = resource;
		}
		
		@Override
		public int compareTo(LogEvent e2) {
			if (time - e2.time < 0){
				return -1;
			}else{
				if (time - e2.time > 0){
					return 1;
				}else{
					System.out.println("the same event");
					return 0;
				}
			}
		}
	}		
}
